TRAININGGROUNDS_GameObject = {}
TRAININGGROUNDS_GameObject.__index = TRAININGGROUNDS_GameObject



function TRAININGGROUNDS_GameObject.new()
	local self=setmetatable({}, TRAININGGROUNDS_GameObject)	
	
	
	------ private vars ------
	
	dying=false		-- object will die at end of frame
	dead=false		-- object is dead, removed from object list, and should be considered invalid by any other objects that still reference it
	
	--------------------------
	
	return self
end

function TRAININGGROUNDS_GameObject:setup(environment)
	
	self.environment=environment
	
	self:SetCustomInfo()
	
	-- default gameobject has no appearance, but we set it up so it'll be visible as soon as a frame is added
	self.parentframe=environment.frame	

	
	self:SetFrame()
	self:SetAppearance()	
	self:SetFrameSize()
	self:SetShadowDetails()
	self:SetShadowFrameIfApplicable()
	
	self:SetClickability()
	self.ismouseovered=false
	
	self.dying=false	-- will be removed at the end of the frame
	self.dead=false		-- removed from gameplay; any object still referencing this one should clear the reference
	
	self.deadingame=false	-- actually "dead"; 0 HP and corpse
	

	--TODO: move to overridable function
	if(self.clickarea~=nil and self.modeloffsetX~=nil and self.modeloffsetY~=nil) then	
		-- clickframeoffset does not affect the position of the clickable area;
		-- it only affects where the frame is drawn on the screen (in case the parent frame is offset, such as with WoWCharacter)
		-- self.clickarea.clickframeoffsetX=-self.modeloffsetX;self.clickarea.clickframeoffsetY=-self.modeloffsetY;
		-- TODO: clickframeoffset appears to be deprecated
		self.clickarea.clickframeoffsetX=0;self.clickarea.clickframeoffsetY=0;
	else
		self.clickarea.clickframeoffsetX=0;self.clickarea.clickframeoffsetY=0;
	end
	
	if(self.frame~=nil) then
		--TODO: set frame level
		self.frame:SetParent(self.parentframe)		
		self:SetDefaultFrameLevel()
		--print("GameObject:setup "..tostring(self.parentframe))
		if(self.clickable==true)then
			self.clickframe=TRAININGGROUNDS_ReusableFrames.getFrame()
			self.clickframe:SetParent(self.frame)
			self:SetDebugVisibleClickareaFrame()
		end
	end
	
	self:SetCombat()
	if(self.combat~=nil)then
		self.combat:CreateNameplate()
		--!!!
		self:SetClickarea()
		self:SetHPToDefault()
	end
	
	
	if(self.visible==true) then
		if(self.frame~=nil)then
			self.frame:Show()
		end

		self:draw()
	end
	
	-- not fully implemented
	self.lockmovementendtick=0
	
	--TODO:
	-- currently only used for WoWCharacter
	-- ...call WoWCharacter:SetExtraScale to change
	self.extrascale=1
	
	--print("GameObject:setup ",self)
	tinsert(environment.newgameobjects,self)
end

function TRAININGGROUNDS_GameObject:SetCustomInfo()
	-- override as necessary
	-- call super when overriding
	self.x=0
	self.y=0
	self.z=0
	self.xspeed=0
	self.yspeed=0
	self.zspeed=0
	self.gravity=0
	self.xaccel=0
	self.yaccel=0
	self.expirytime=nil
	
	-- some of these can be overridden via SetAppearance
	-- we initialize them here so they don't have to be explicitly mentioned in SetAppearance
	-- (which usually doesn't call child function when overridden)
	self.alignment="CENTER"
	
	-- TODO: visible is not yet implemented anywhere except during setup
	self.visible=true
	
	self.ysortable=true	
	
	-- Can Go OOB
	-- (better not set this until the object is defeated)
	self.cangooob=false
	
	self.defaulthp=nil
end

function TRAININGGROUNDS_GameObject:SetClickability()
	-- override. don't call super when overriding
	-- setup function will create a debug-visible clickframe if clickable==true AND frame is visible and not nil.
	-- for various convenience reasons, clickarea is setpoint(bottom, center) instead of center/center or bottomleft etc.
	
	-- remember to call SetClickability again if you ever resize an object after setup!
	
	self.clickable=false	
	--self.clickable=true
	self.clickarea={}
	self.clickarea.x=0*4; self.clickarea.y=0*4; self.clickarea.width=15*4; self.clickarea.height=15*4

end

function TRAININGGROUNDS_GameObject:SetDebugVisibleClickareaFrame()
	-- This controls the visible clickframe only
	-- ...the actual dimensions are controlled by SetClickarea/SetClickability
		-- TODO: if base GameObject class doesn't use SetClickarea, start using it -- looks like it's only in WoWCharacter
	if(self.clickable==true and self.clickframe~=nil and _TRAININGGROUNDS.debug_showclickframes==true) then
		local camera=self.environment.camera
		--don't apply resolutioncorrection to clickframe size
		--local resolutioncorrection=((GetScreenWidth()/GetScreenHeight())/(800.0/600.0))   --/self.parentframe:GetScale()
		--print("TRAININGGROUNDS_GameObject:SetClickFrameSize " .. resolutioncorrection)

		self.clickframe:SetParent(self.frame)
		
		self.clickframe:Show()
		self.clickframe.texture:SetTexture(1,1,1,.5)			
		self.clickframe.texture:Show()
			
		--TODO: "self.modelframe" doesn't exist for GameObjects
		local adjustedzoom=camera.zoom*self.modelinfo.gameplayscaleratio*self.extrascale
		self.clickframe:SetPoint("BOTTOM",self.frame,"CENTER",(self.clickarea.x)*adjustedzoom,(self.clickarea.y)*adjustedzoom)
		
		local width=self.clickarea.width*adjustedzoom
		local height=self.clickarea.height*adjustedzoom
		
		self.clickframe:SetWidth(width)
		self.clickframe:SetHeight(height)	
		self.clickframe:SetFrameLevel(255)
	end
end

function TRAININGGROUNDS_GameObject:SetFrame()
	-- override as necessary
	-- beware when calling child functions; frames don't like being set twice
	-- TODO: create frame as default behavior? (especially since we seem to expect frame to exist at some other points)
												-- (add more checks to make sure frame exists)
	
	self.frame=nil
end

function TRAININGGROUNDS_GameObject:SetAppearance()
	-- override as necessary
	-- beware when calling child functions; modelframes don't like being set twice
end

function TRAININGGROUNDS_GameObject:SetFrameSize()
	-- override as necessary
	-- beware when calling child functions; frames don't like being set twice
	
end

function TRAININGGROUNDS_GameObject:SetShadowDetails()
	-- override as necessary
	self.shadow=false
end

function TRAININGGROUNDS_GameObject:SetShadowFrameIfApplicable()
	self.shadowframe=nil
	if(self.shadow==true)then
		self.shadowframe=TRAININGGROUNDS_ReusableFrames.getFrame()
		
		-- --self.shadowframe:SetParent(self.parentframe)
		-- self.shadowframe:SetParent(UIParent)
		-- self.shadowframe:SetWidth(200);self.shadowframe:SetWidth(200) --!!!
		-- self.shadowframe:SetPoint("BOTTOMLEFT",UIParent,"BOTTOMLEFT",100,50)
		-- self.shadowframe:SetFrameLevel(200)
		-- self.shadowframe:Show()
		-- self.shadowframe.texture:Show()
		-- self.shadowframe.texture:SetTexture(1,1,1,1)
		
		--self.shadowframe:SetParent(self.frame)		
		self.shadowframe:SetParent(self.parentframe)
		local width=100
		--TODO: tidy up
		if(self.modelinfo~=nil) then			
			--width=self.modelinfo.width
			width=self.modelinfo.clickwidth
		end
		self.shadowwidth=width
		self.shadowframe:SetWidth(width*self.environment.camera.zoom)
		--print("Shadow width: ",width*self.environment.camera.zoom)
		-- TODO: calculate shadow height the same way we calculate clickarea (if we're not already doing that now)
		self.shadowframe:SetHeight(width*self.environment.camera.ystretch*self.environment.camera.zoom)
		self.shadowframe:SetPoint("CENTER",0,0)
		self.shadowframe:SetFrameLevel(self.environment.shadowframelevel)
		--self.shadowframe:SetFrameLevel(self.environment.shadowframelevel)
		self.shadowframe.texture=self.shadowframe:CreateTexture()
		self.shadowframe.texture:SetAllPoints(self.shadowframe)
		self.shadowframe.texture:SetTexture(0,0,0,1)				
		self.shadowframe.texture:Show()
		self.shadowframe:Show()		
		self.shadowframe.texture:SetTexture("Interface\\AddOns\\TrainingGrounds\\Assets\\Graphics\\shadow.tga")		
		-- self.shadowframe.texture:SetTexture("Interface\\AddOns\\TrainingGrounds\\Assets\\Graphics\\targetingcircle.tga")
		-- self.shadowframe.texture:SetGradientAlpha("HORIZONTAL",1,0,0,1.0,1,0,0,1.0);
		
	end
end

function TRAININGGROUNDS_GameObject:SetCombat()
	-- override
	-- do not call super when overriding
end

-- TODO: check if no longer in use?
function TRAININGGROUNDS_GameObject:SetNameplate()
	-- override
	-- do not call super when overriding
end

function TRAININGGROUNDS_GameObject:SetDefaultFrameLevel()
		--TODO: determine if we want the default frame level to be 255 or not
		--	-- answer may change depending on whether we keep ysort-every-several-frames or not
	self.frame:SetFrameLevel(255)
end

function TRAININGGROUNDS_GameObject:step(elapsed)
	
	local currenttime=self.environment.currenttime
	

	
	
	
	
	--TODO: move movement to a separate function that only loops after entire AI loop is done
	
	self.xspeed=self.xspeed+self.xaccel*elapsed
	self.yspeed=self.yspeed+self.yaccel*elapsed
	
	self.x=self.x+self.xspeed*elapsed
	self.y=self.y+self.yspeed*elapsed
	
	self.prevz=self.z
	

	
	self.z=self.z+self.zspeed*elapsed
	self.zspeed=self.zspeed+self.gravity*elapsed
	if(self.z<=0) then
		if(self.prevz>0) then
			self:OnLandOnGround()
		else
			self:BelowGround()
		end
	end	
	
	
	if(self.expirytime~=nil)then
		if(currenttime>=self.expirytime) then		
			self:die()
		end
	end
end


function TRAININGGROUNDS_GameObject:OnLandOnGround()
	--don't call super when overriding
	self:BelowGround()
	-- self.z=math.abs(self.z)
	-- self.zspeed=math.abs(self.zspeed)/2
end

function TRAININGGROUNDS_GameObject:BelowGround()
	--TODO: call super?	
	self.z=0
	self.zspeed=0
end




function TRAININGGROUNDS_GameObject:animate()
	--override
end

function TRAININGGROUNDS_GameObject:draw()
	
	if(self.frame~=nil) then
	
		local camera=self.environment.camera
		
		if(camera.zoomchanged==true)then
			self:SetFrameSize()	--also calls SetClickFrameSize
			--self.nameplate:ResetPoint()
		end
		
		

		
		--local scaledoffsetX=0
		--local scaledoffsetY=0
		--self.modelframe:SetPoint("CENTER",self.frame,"CENTER",scaledoffsetX,scaledoffsetY)
		local scrollxoffset,scrollyoffset--=camera:GetScrollOffset()
		scrollxoffset=0;scrollyoffset=0 	--scrolling effect is now handled by scrollframe
		local x=(self.x+scrollxoffset)*camera.zoom
		local y=((self.y+scrollyoffset)*camera.zoom*camera.ystretch)
		local TEMP_screenscale=1	--1.5
		local zoffset=(self.z*camera.zoom*camera.zstretch*TEMP_screenscale)	
		--self.frame:SetPoint("CENTER",self.parentframe,"BOTTOMLEFT",x,y)
		--self.frame:SetPoint("CENTER",self.parentframe,"CENTER",x,y+zoffset)
		--self.frame:SetPoint(camera.objectanchor,self.parentframe,"CENTER",x,y+zoffset)
		--self.frame:SetPoint(camera.objectanchor,self.parentframe,self.alignment,x,y+zoffset)
		
		-- SetPoint: OWN object's point, parent, PARENT'S point, x, y
		-- if you change object's alignment after setup, remember to ClearAllPoints
		self.frame:SetPoint(self.alignment,self.parentframe,camera.objectanchor,x,y+zoffset)
		--self.frame:SetPoint("CENTER",self.parentframe,"CENTER",x,y+zoffset)
	
	end
	
	--print("gameobject draw "..x..", "..y+zoffset..", "..self.alignment..", "..camera.objectanchor)
	
end


function TRAININGGROUNDS_GameObject:SetHPToDefault()
	if(self.defaulthp~=nil)then
		self.combat:SetMaximumAndCurrentHP(self.defaulthp)
	end
end

-- for now, melee range is distance<=50
function TRAININGGROUNDS_GameObject:HitboxDistanceFromOtherObject(obj)
	local dist=math.sqrt((self.x-obj.x)*(self.x-obj.x)+(self.y-obj.y)*(self.y-obj.y))
	--dist=dist-self.modelinfo.width/2
	--dist=dist-obj.modelinfo.width/2

	dist=dist-self:HitboxRadius()
	dist=dist-obj:HitboxRadius()

	return dist
end

function TRAININGGROUNDS_GameObject:HitboxRadius()
	if(self.modelinfo==nil or self.modelinfo.clickwidth==nil)then
		return 0
	end
	return self.modelinfo.clickwidth/2
end

function TRAININGGROUNDS_GameObject:SetTargetingCircleColor(player)
	if(player.combat.target~=self)then return end
	if(self.deadingame==true)then
		self.shadowframe.texture:SetGradientAlpha("HORIZONTAL",0.5,0.5,0.5,0.5,0.5,0.5,0.5,0.5);
		return
	end
	if(self==player)then
		self.shadowframe.texture:SetGradientAlpha("HORIZONTAL",0,1,0,1.0,0,1,0,1.0);
	else
		self.shadowframe.texture:SetGradientAlpha("HORIZONTAL",1,0,0,1.0,1,0,0,1.0);
	end
end

function TRAININGGROUNDS_GameObject:die()
	--print("GameObject:die")
	self.dying=true
end

function TRAININGGROUNDS_GameObject:Cleanup()
	self.dead=true
	--print("GameObject:cleanup 1")
	if(self.frame~=nil)then
		--print("GameObject:cleanup 2")
		if(self.frame.Cleanup~=nil)then		
			--print("GameObject:cleanup 3")
			self.frame.Cleanup(self.frame)	
			-- N.B.: we don't call frame:Cleanup() because we're trying not to treat frame as OOP
			--TODO: ...maybe start treating it as OOP
		end
	end
	if(self.clickframe~=nil)then
		if(self.clickframe.Cleanup~=nil)then
			self.clickframe:Cleanup()
		end
	end
	if(self.shadowframe~=nil)then
		self.shadowframe:Cleanup()
	end
	if(self.combat~=nil)then
		self.combat:Cleanup()
	end
end
