TRAININGGROUNDS_ReusableFrames = {}

TRAININGGROUNDS_ReusableFrames.inactiveframes={}
TRAININGGROUNDS_ReusableFrames.activeframes={}

function TRAININGGROUNDS_ReusableFrames.removeFrame(f)
	--print("Running cleanup on ",f.name)
	
	f:Hide()
	f:ClearAllPoints()	
	f.texture:SetTexture(0,1,1,0.5)
	f.texture:SetTexCoord(0,1,0,1)
	f.texture:SetGradientAlpha("HORIZONTAL",1,1,1,1,1,1,1,1)
	f.texture:Hide()
	f.fontstring:Hide()
	f.fontstring:SetFont("Fonts\\FRIZQT__.TTF",14,"OUTLINE")
	f.fontstring:SetJustifyH("CENTER")
	f.fontstring:SetTextHeight(14)	
	f.framelevel=0	
	f:SetFrameLevel(f.framelevel)	
	--TODO: copy to other reusableframe files
	--f:SetParent(nil)
	f:SetParent(UIParent)
	if(f:GetParent()~=UIParent)then
		print("WARNING: after cleanup, frame "..tostring(f:GetName()).."'s parent is not UIParent but "..tostring(f:GetParent():GetName()).." instead")
	end

	if(f.animgroup~=nil) then
		f.animgroup:Stop()
		f.animgroup=nil
	end
	if(f.texture.animgroup~=nil)then
		f.texture.animgroup:Stop()
		f.texture.animgroup=nil
	end
	f:SetAlpha(1)
	
	_TRAININGGROUNDS:tremovebyval(TRAININGGROUNDS_ReusableFrames.activeframes,f)	
	if(TRAININGGROUNDS_tcontains(TRAININGGROUNDS_ReusableFrames.inactiveframes,f))then
		local errorstring="TRAININGGROUNDS_ReusableFrames.inactiveframes already contains "..tostring(f).." "..f.name.." "..f.tag
		print("WARNING: "..errorstring)
		--error(errorstring)
	else
		tinsert(TRAININGGROUNDS_ReusableFrames.inactiveframes,f)
	end
	
end

TRAININGGROUNDS_ReusableFrames.nextframeid=0
function TRAININGGROUNDS_ReusableFrames.getFrame()
	local f=tremove(TRAININGGROUNDS_ReusableFrames.inactiveframes)
	if not f then
		local framename="TRAININGGROUNDS_ReusableFrame"..tostring(TRAININGGROUNDS_ReusableFrames.nextframeid);
		f=CreateFrame("Frame",framename,UIParent)
		f.id=TRAININGGROUNDS_ReusableFrames.nextframeid
		f.name=framename
		--TODO: add tag to the rest of the reusableframe family
		f.tag="(no tag)"
		f.framelevel=0
		f:SetFrameLevel(f.framelevel)		
		f:Hide()
		f.texture = f:CreateTexture()
		f.texture:SetAllPoints(f)
		f.texture:SetTexture(0,1,1,0.5)	
		f.texture:SetGradientAlpha("HORIZONTAL",1,1,1,1,1,1,1,1);
		f.texture:Hide()		
		f.fontstring=f:CreateFontString(nil, "ARTWORK")
		f.fontstring:SetAllPoints()
		f.fontstring:Hide()
		f.fontstring:SetFont("Fonts\\FRIZQT__.TTF",14,"OUTLINE")
		f.fontstring:SetTextHeight(14)	
		f.animgroup=nil
		-- most gameobjects expect to have a "self.frame", but they don't know what type of frame that is
		-- (could be normal frame, buttonframe, modelframe, scrollframe, etc)
		-- so we'll store a reference to the cleanup function in the frame itself
		f.Cleanup=TRAININGGROUNDS_ReusableFrames.removeFrame
		
		TRAININGGROUNDS_ReusableFrames.nextframeid=TRAININGGROUNDS_ReusableFrames.nextframeid+1
	else
		--
		if(f:GetParent()~=UIParent)then
			print("WARNING: frame "..tostring(f:GetName()).."'s parent is not UIParent but "..tostring(f:GetParent():GetName()).." instead")
		end
	end
	--print("ReusableFrames getting a frame ",f.name)
	tinsert(TRAININGGROUNDS_ReusableFrames.activeframes,f)

	return f
end

function TRAININGGROUNDS_ReusableFrames.resetAllInactiveFrames()
	local looping=true
	while looping==true do
		looping=false
		local f
		f=nil
		if(#TRAININGGROUNDS_ReusableFrames.activeframes>0) then			
			f=TRAININGGROUNDS_ReusableFrames.activeframes[1]
		end
		if not f then
			--
		else
			TRAININGGROUNDS_ReusableFrames.removeFrame(f)
			if(f:GetParent()~=UIParent)then
				--print("WARNING: during force cleanup, frame "..tostring(f:GetName()).."'s parent is not UIParent but "..tostring(f:GetParent():GetName()).." instead")
			end
			--print("Force cleanup on ReusableFrames: "..f:GetName().." cleaned up, "..#TRAININGGROUNDS_ReusableFrames.activeframes.." frames remaining")
			looping=true
		end
	end
end

function TRAININGGROUNDS_ReusableFrames.ResetAndReport()
	local oktotal=#TRAININGGROUNDS_ReusableFrames.inactiveframes
				+#TRAININGGROUNDS_ReusableModelFrames.inactiveframes
				+#TRAININGGROUNDS_ReusableParticleFrames.inactiveframes
				+#TRAININGGROUNDS_ReusableButtonFrames.inactiveframes				
				+#TRAININGGROUNDS_ReusableScrollFrames.inactiveframes
	local forcetotal=#TRAININGGROUNDS_ReusableFrames.activeframes
				+#TRAININGGROUNDS_ReusableModelFrames.activeframes
				+#TRAININGGROUNDS_ReusableParticleFrames.activeframes
				+#TRAININGGROUNDS_ReusableButtonFrames.activeframes
				+#TRAININGGROUNDS_ReusableScrollFrames.activeframes
	--_TRAININGGROUNDS:debug_print("Properly cleaned up "..tostring(oktotal).." frames.  Running force cleanup on "..tostring(forcetotal).." frames.")
	
	--TODO: rename to ResetAllActiveFrames
	TRAININGGROUNDS_ReusableFrames.resetAllInactiveFrames()
	TRAININGGROUNDS_ReusableModelFrames.resetAllInactiveFrames()
	TRAININGGROUNDS_ReusableParticleFrames.resetAllInactiveFrames()
	TRAININGGROUNDS_ReusableButtonFrames.resetAllInactiveFrames()
	TRAININGGROUNDS_ReusableScrollFrames.resetAllInactiveFrames()
	
end

function TRAININGGROUNDS_ReusableFrames.Report()
	print("Checking ReusableFrames only...")
	print(tonumber(#TRAININGGROUNDS_ReusableFrames.activeframes).." active frames")
	print(tonumber(#TRAININGGROUNDS_ReusableFrames.inactiveframes).." inactive frames")
	for i=1,#TRAININGGROUNDS_ReusableFrames.inactiveframes,1 do
		local f=TRAININGGROUNDS_ReusableFrames.inactiveframes[i]
		if(f:GetParent()~=UIParent)then
			--print("Frame "..tostring(f:GetName()).."'s parent is not UIParent but "..tostring(f:GetParent():GetName()).." instead")
		end
		--print("Frame "..tostring(f:GetName()).."'s parent is "..tostring(f:GetParent():GetName()))
	end
end